#########################################################################
#
# playedit for tk3play 2.0a last edit: 03/01/98
#  
#  Copyright (C) 1997-1998 Brian Foutz (bef2@cornell.edu)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#
########################################################################

set PlayEdit(listfont) -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*
set PlayEdit(entryfont) -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*
set PlayEdit(bgcolor) gray65
set PlayEdit(fgcolor) black 
set PlayEdit(fghighlight) gray30 
set PlayEdit(bghighlight) midnightblue 
set PlayEdit(crcolor) white

set PlayEdit(fchoices) {{*.mp*} {*} {*.mp2} {*.mp3}}

set IMGDIR /usr/local/lib/tk3play-2.0b


#
# FileList procedures
#

proc FileList(select) { } {
   global FileList
   set sel [$FileList(filelist) curselection]
   set first [lindex $sel 0]
   set FileList(filename) [$FileList(filelist) get $first]
#   set FileList(filename) [$FileList(filelist) get active]
}

proc FileList(enter) { } {
   global FileList DirList

   if {![check_file $DirList(directory)$FileList(filename)]} {return}

   # Find entered filename in the filelist
   set index [lsearch [$FileList(filelist) get 0 end] $FileList(filename)]
   $FileList(filelist) selection clear 0 end
   $FileList(filelist) selection set $index
   $FileList(filelist) selection anchor $index
   return
}

proc FileList(update) {dirname} {
   global FileList

   $FileList(filelist) delete 0 end
   foreach i [lsort [glob -nocomplain $dirname/$FileList(filter) \
                                      $dirname/.$FileList(filter)]] {
      if {![file isdirectory $i]} {
         $FileList(filelist) insert end [file tail $i]
      }
   }

   # blank out entry window
   set FileList(filename) ""
   return
}

#
# DirList procedures
#

proc DirList(select) { } {
   global DirList
   set sel [$DirList(dirlist) curselection]
   set dirname [lindex $DirList(dirname_list) $sel]
   set DirList(directory) $dirname
   DirList(update) $dirname
   return
}

proc DirList(append) { x y } {
   global DirList PlayList

   set sel [$DirList(dirlist) index @$x,$y]
   set dirname [lindex $DirList(dirname_list) $sel]
   lappend PlayList(playlist) "$dirname*.mp*"
   if {$PlayList(filename) == ""} {
      set t1 [file dirname $PlayList(playlist)]
      set t2 [file tail $t1]
      set PlayList(filename) $t2
   }
   PlayList(update)
   return
}

proc DirList(update) {dirname} {
   global FileList DirList
   global status

   # Fix beginning of dirname
   if {[string index $dirname 0] != "/"} {
      # prepend pwd
      set dirname [pwd]/$dirname
   }

   if {![check_dir $dirname]} {return}

   # Fix end of dirname
   set len [expr [string length $dirname] - 1]
   if {[string index $dirname $len] == "/"} {
      # Remove final "/"
      set dirname [string range $dirname 0 [expr $len - 1]]
   }
   # Now all dirnames are full paths with no / at the end
   # (so we can split it properly)

   $DirList(dirlist) delete 0 end
   set dir_tokens [split $dirname "/"]
   if {$dir_tokens == ""} {
      set dir_tokens "{}"
   }
   set offset ""
   set DirList(dirname_list) ""
   set last_dir ""
   foreach i $dir_tokens {
      set s $offset$i/
      $DirList(dirlist) insert end $s
      append offset "  "
      append last_dir $i/
      lappend DirList(dirname_list) $last_dir
   }

   foreach i [lsort [glob -nocomplain $dirname/* $dirname/.*]] {
      if {[file isdirectory $i]} {
         set s [file tail $i]/
         if {$s != "./" && $s != "../"} {
            $DirList(dirlist) insert end $offset$s
            lappend DirList(dirname_list) $last_dir$s
         }
      }
   }

   set len [llength $dir_tokens]
   $DirList(dirlist) selection set [expr $len - 1]
   set DirList(directory) $dirname/

   FileList(update) $dirname
   return
}

proc check_file {filename} {
   global status

   set file [glob -nocomplain $filename]
   if {[llength $file] != 1} {
      set status "$filename is not a valid file"
      return 0
   } elseif {![file isfile $filename]} {
      set status "$filename is not a valid file"
      return 0
   } else {
      set status ""
      return 1
   }
}

proc check_dir {dirname} {
   global status

   set dir [glob -nocomplain $dirname]
   if {[llength $dir] != 1} {
      set status "$dirname is not a valid directory"
      return 0
   } elseif {![file isdirectory $dirname]} {
      set status "$dirname is not a valid directory"
      return 0
   } else {
      set status ""
      return 1
   }
}

proc PlayList(clear) {} {
   global PlayList status

   if {[info exists PlayList(playlist)]} {unset PlayList(playlist)}
   set PlayList(filename) ""
   PlayList(update)
   set status ""
}

proc PlayList(remove) {} {
   global PlayList

   # Remove all selected elements from playlist
   set sel [$PlayList(listbox) curselection]
   set sorted [lsort -decreasing -integer $sel]
   foreach i $sorted {
      set PlayList(playlist) [lreplace $PlayList(playlist) $i $i]
   }
   PlayList(update)
   return
}

proc PlayList(update) {} {
   global PlayList status

   $PlayList(listbox) delete 0 end
   if {[info exists PlayList(playlist)]} {
      foreach i $PlayList(playlist) {
         set len [$PlayList(listbox) size]
         set next [expr $len + 1]
         set space "    "
         set s $next.$space$i
         $PlayList(listbox) insert end $s
      }
   }
   set status ""
   return
}

proc PlayList(append) {} {
   global PlayList
   global FileList DirList

   # Append elements from FileList
   set sel [$FileList(filelist) curselection]
   foreach i $sel {
      set file [$FileList(filelist) get $i]
      set file $DirList(directory)$file
      if {[file extension $file] == ".mp3" || 
          [file extension $file] == ".mp2" || 
          [file extension $file] == ".MP3" ||
          [file extension $file] == ".MP2"} {
         lappend PlayList(playlist) $file
      } else {
         # Ask user about playlist selection
#         set button [tk_dialog .query "Append Playlist?" \
#            "Do you want to append the contents of the playlist $file?" \
#            "" 0 Yes Cancel]
#         if {$button == 0} {
            # Load the contents of the playlist file (1 = append mode)
            PlayList(load) $file 1
#         }
      }
   }
   PlayList(update)
   $PlayList(listbox) see end
   return
}

proc PlayList(replace) { } {
   global PlayList FileList

   set numsel [llength [$FileList(filelist) curselection]]
   if {[info exists PlayList(playlist)]} {
      set button [tk_dialog .query "Replace Playlist?" \
      "Do you want to replace the playlist with the $numsel selected files?" \
         "" 0 Yes Cancel]
      if {$button != 0} {return}
   }
   PlayList(clear)
   PlayList(append)

   return
}

proc PlayList(load) {playlist {append 0}} {
   global PlayList
   global FileList
   global status

   if {!$append} {
      PlayList(clear)
   }
   set fp [open $playlist r]
   set good 1
   while {![eof $fp] && $good} {
      set playline [gets $fp]
      set length [llength $playline]
      if {$length > 1} {
         if {[scan "$playline" "%d" pause] == 1} {
	    # Pause value present
	    # Scan again to find the starting index of the filename
            if {[scan $playline "%d %s" spause string1] != 2} {
               set status "error in playlist $listfile : line = $playline"
               set good 0
            }
            set starti [string first $string1 "$playline"]
	    set filespec [string range "$playline" $starti end]
         } else {
	    # Set filespec = entire line (w/ spaces)
 	    set filespec "$playline"
         }
      } elseif {$length == 1} {
         # Set filespec = entire line (no spaces)
         set filespec "$playline"
      } else {
         # length = 0  -blank line -skip
         continue
      }

      lappend PlayList(playlist) $filespec
   }
   close $fp
   set PlayList(filename) $playlist
   PlayList(update)
   return
}

proc PlayList(save) {} {
   global PlayList PLAYDIR
   global status

   if {$PlayList(filename) == ""} {
      set status "Playlist filename is blank"
      return
   }

   if {[string index $PlayList(filename) 0] != "/"} {
      # Save to $PLAYDIR
      set PlayList(filename) $PLAYDIR/$PlayList(filename)
   }
   set fp [open $PlayList(filename) w]
   foreach i $PlayList(playlist) {puts $fp $i}
   close $fp
   set status "Playlist saved in file $PlayList(filename)"
   return
}

proc PlayList(saveas) {} {
   global PlayList
   global status
   set playlist [tk_getSaveFile -defaultextension m3u \
      -title "Save PlayList As"]
   if {$playlist != ""} {
      set status "Saving to $playlist"
      set PlayList(filename) $playlist
      PlayList(save)
   }
   return
}

proc PlayList(open) {} {
   global PlayList status
   set playlist [tk_getOpenFile -defaultextension m3u -title "Open Playlist"]
   if {$playlist != ""} {
      PlayList(load) $playlist
   }
   set status ""
   return
}

proc debracket {filespec} {
   set i [string first "\[" $filespec]
   while {$i != -1} {
     set filespec [string range $filespec 0 [expr $i - 1]]?[string range $filespec [expr $i + 1] end]
     set i [string first "\[" $filespec]
   }

   set i [string first "\]" $filespec]
   while {$i != -1} {
     set filespec [string range $filespec 0 [expr $i - 1]]?[string range $filespec [expr $i + 1] end]
     set i [string first "\]" $filespec]
   }

   return $filespec
}

proc PlayList(apply) {} {
   global PlayList status cdir
   upvar #0 state s
   upvar #0 playlist p
   upvar #0 display d
   upvar #0 temp t

   set t(total_time) 0
   set t(playlist) ""

   if {![info exists PlayList(playlist)]} { return }
   foreach filespec $PlayList(playlist) {
      set filespec [debracket $filespec]
      set filelist [lsort [glob -nocomplain "$filespec"]]
      if {[llength $filelist] == 0} {
         set status "$filespec matches no files"
      } else {
         foreach filename $filelist {
            if {[file isfile "$filename"]} {
               playlist(add) "$filename" $s(bgpause) temp
            } else {
	       if {$filename != "."} {
                  set status "$filename is an invalid file"
               }
	    }
	 }
      }
   }

   if {$t(total_time) != 0} {
      set c [playlist(curr_track)]
      if {$p(currIndex) != -1} {
        if {[llength $t(playlist)] < [expr $c + 1]} {
           if {$s(pipe_open)} {
              stop
           }
           set s(csecs) 0
           set p(currIndex) 0
        } elseif {[playlist(curr_song)] != [lindex $t(playlist) $c]} {
           if {$s(pipe_open)} {
              stop
           }
           set s(csecs) 0
           set p(currIndex) 0
        }
      } else {
        set s(csecs) 0
        set p(currIndex) 0
      }

      # Copy the temp playlist into the real one.
      set p(songlen_list) $t(songlen_list)
      set p(total_time) $t(total_time)
      set p(layer_list) $t(layer_list)
      set p(outrate_list) $t(outrate_list)
      set p(inrate_list) $t(inrate_list)
      set p(mode_list) $t(mode_list)
      set p(playlist) $t(playlist)
      set p(pause_list) $t(pause_list)
      set p(mpegnum_list) $t(mpegnum_list)
      # Clear temp playlist
      unset t(songlen_list)
      unset t(total_time)
      unset t(layer_list)
      unset t(outrate_list)
      unset t(inrate_list)
      unset t(mode_list)
      unset t(playlist)
      unset t(pause_list)
      unset t(mpegnum_list)
#      puts "near end"
#      if {$PlayList(filename) != ""} {
         set d(pl_name) [file tail $PlayList(filename)]
#         set d(pl_name) [string toupper [file tail $PlayList(filename)]]
#      }
      if {$p(currIndex) == -1} {set p(currIndex) 0}
      playlist(shuffle)
      goto_fill
      init_display
   } else {
#      puts "Playlist $listfile contained invalid entries"
   }
   return
}

proc modal_edit { } {
   global PlayList update playline pos

   set sel [$PlayList(listbox) curselection]
   if {[llength $sel] == 0} {return}
   set orig [lindex $sel 0]
   set playline [lindex $PlayList(playlist) $orig]

   set update 0
   toplevel .edit
   wm title .edit "Edit Playlist Line"
   set pos [expr $orig + 1]
   frame .edit.1
   label .edit.1.label -text "Playlist position:"
   entry .edit.1.pos -textvariable pos -width 4
   frame .edit.2
   label .edit.2.label -text "Filename:"
   entry .edit.2.entry -textvariable playline -width 50
   button .edit.ok -text "OK" -command {set update yes}
   button .edit.cancel -text "Cancel" -command {set update no}
   pack .edit.1 -side top -fill x -expand 1
   pack .edit.1.label -side left -anchor w
   pack .edit.1.pos -side left -anchor w
   pack .edit.2 -side top -fill x -expand 1
   pack .edit.2.label -side left -anchor w
   pack .edit.2.entry -side left -anchor w -fill x
   pack .edit.ok -side left -expand 1
   pack .edit.cancel -side left -expand 1

   tkwait variable update

   # Update the playlist be deleting the current
   # entry and putting the new one in.
   if {$update == "yes"} {
      set PlayList(playlist) [lreplace $PlayList(playlist) $orig $orig]
      set new [expr $pos - 1]
      set PlayList(playlist) [linsert $PlayList(playlist) $new $playline]
   }
   destroy .edit
   PlayList(update)
   return
}

#
# Remove the window
#
proc PlayList(exit) { } {
   global .playedit cdir DirList
   upvar #0 state s

   set cdir $DirList(directory)
   destroy .playedit
   set s(playedit_open) 0
   return
}

#
# Create the window
#
proc PlayList(playedit) { } {
   global status FileList DirList FileList PlayList PlayEdit
   global PLAYDIR version cdir IMGDIR
   upvar #0 state s
   upvar #0 playlist p

   set FileList(fchoices) $PlayEdit(fchoices)
   set FileList(filter) "*"
   set status ""

   if {$s(playedit_open)} {return}
   if {![info exists cdir]} {
      if {[file exists $PLAYDIR]} {
         set cdir [glob $PLAYDIR]
      } else {
         set cdir [pwd]
      }
   }

   set s(playedit_open) 1

   # Create the playedit window
   toplevel .playedit
   wm title .playedit "Playlist Editor"

   # Status bar on the bottom

   label .playedit.b -textvariable status -relief sunken -anchor w
   pack .playedit.b -side bottom -fill x


   #buttons on top

   frame .playedit.mb
   set b .playedit.mb
   button $b.clear -image [image create photo -file $IMGDIR/clear.gif] -command {PlayList(clear)}
   button $b.edit -image [image create photo -file $IMGDIR/edit.gif] -command {modal_edit} 
   button $b.replace -image [image create photo -file $IMGDIR/replace.gif] -command {PlayList(replace)}
   button $b.append -image [image create photo -file $IMGDIR/append.gif] -command {PlayList(append)}
   button $b.remove -image [image create photo -file $IMGDIR/remove.gif] -command {PlayList(remove)}
   button $b.open -image [image create photo -file $IMGDIR/open.gif] -command {PlayList(open)} 
   button $b.save -image [image create photo -file $IMGDIR/todisk.gif] -command {PlayList(save)}
   button $b.saveas -image [image create photo -file $IMGDIR/saveas.gif] -command {PlayList(saveas)}
   button $b.play -image [image create photo -file $IMGDIR/totk3play.gif] -command {PlayList(apply)}
   button $b.close -image [image create photo -file $IMGDIR/close.gif] -command {PlayList(exit)}
   pack $b -side top -fill x
   pack $b.close -side right
   pack $b.open -side right 
   pack $b.save -side right
   pack $b.saveas -side right
   pack $b.play -side right
   pack $b.clear -side right 
   pack $b.edit  -side right
   pack $b.replace -side right
   pack $b.remove -side right
   pack $b.append -side right



   # Directory Stuff

   frame .playedit.t
   set d .playedit.t.d
   frame $d -relief flat -bd 2 
   label $d.label -text "Directory:" -bg grey40 -fg white -font  -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-* -relief sunken
   entry $d.entry -textvariable DirList(directory) -relief sunken \
      -width 35 -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-* -bg $PlayEdit(bgcolor) \
      -fg $PlayEdit(fghighlight) -insertbackground $PlayEdit(crcolor)
   set DirList(entry) $d.entry
   frame $d.m -relief sunken -bd 2
   set DirList(dirlist) $d.m.dirlist
   scrollbar $d.m.scroller -command "$d.m.dirlist yview" -width 16
   listbox $d.m.dirlist -relief flat \
      -yscrollcommand "$d.m.scroller set" -bg $PlayEdit(bgcolor) \
      -fg $PlayEdit(fgcolor) -selectbackground $PlayEdit(bgcolor) \
      -selectforeground $PlayEdit(fghighlight) -selectborderwidth 0 \
      -font $PlayEdit(listfont)

   pack .playedit.t -side top -fill both -expand 1
   pack $d -side left -fill both -expand 1
   pack $d.label -side top -fill x -anchor w
   pack $d.entry -side top -fill x -anchor w
   pack $d.m -side top -fill both -expand 1
   pack $d.m.scroller -side left -fill y
   pack $d.m.dirlist -side left -fill both -expand 1

   bind $d.entry <Key-Return> {
      if {[check_dir $DirList(directory)] == 1} {
        DirList(update) $DirList(directory) }
   }
   bind $d.m.dirlist <ButtonRelease-1> {DirList(select)}
   bind $d.m.dirlist <ButtonRelease-3> {DirList(append) %x %y}

   # File Stuff

   set f .playedit.t.f
   frame $f -relief flat -bd 2
   label $f.filelab -text "Filename:" -bg grey40 -fg white -font  -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-* -relief sunken
   entry $f.filename -textvariable FileList(filename) -relief sunken \
      -width 35 -font $PlayEdit(entryfont) \
      -bg $PlayEdit(bgcolor) -fg $PlayEdit(fghighlight) \
      -highlightcolor $PlayEdit(bgcolor) -insertbackground $PlayEdit(crcolor)
   set FileList(entry) $f.filename
   frame $f.m -relief sunken -bd 2
   listbox $f.m.filelist -relief flat \
      -yscrollcommand "$f.m.scroller set" -bg $PlayEdit(bgcolor) \
      -fg $PlayEdit(fgcolor) -selectbackground $PlayEdit(bgcolor) \
      -selectforeground $PlayEdit(fghighlight) -font $PlayEdit(listfont) \
      -selectborderwidth 0 -height 10 -selectmode extended
   set FileList(filelist) $f.m.filelist
   scrollbar $f.m.scroller -command "$FileList(filelist) yview" -width 16
   frame $f.ff
   label $f.ff.filtlab -text "Filter:" -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*
   entry $f.ff.filter -textvariable FileList(filter) -relief sunken \
      -font $PlayEdit(entryfont) -bg $PlayEdit(bgcolor) \
      -fg $PlayEdit(fghighlight) -insertbackground $PlayEdit(crcolor)
   set FileList(filter_entry) $f.ff.filter
   menubutton $f.ff.choice -text "Choice" -relief raised -indicatoron 1 \
      -menu $f.ff.choice.menu
   set FileList(choice) $f.ff.choice
   menu $f.ff.choice.menu
   foreach i $FileList(fchoices) {
      $f.ff.choice.menu add command -label "$i" -command \
         "set FileList(filter) $i; FileList(update) \$DirList(directory)"
   }
   pack $f -side right -fill both -expand 1
   pack $f.filelab -side top -fill x
   pack $f.filename -side top -fill x
   pack $f.ff -side bottom -fill x
   pack $f.ff.filtlab -side left -anchor w
   pack $f.ff.choice -side right
   pack $f.ff.filter -side left -fill x -expand 1
   pack $f.m.scroller -side left -fill y
   pack $f.m -side top -fill both -expand 1 -anchor w
   pack $f.m.filelist -side left -fill both -expand 1

   bind $f.filename <Key-Return> {FileList(enter)}
   bind $f.m.filelist <ButtonRelease-1> {FileList(select)}
   bind $f.m.filelist <Double-1> {PlayList(append)}
   bind $f.ff.filter <Key-Return> {FileList(update) $DirList(directory)}

   # The middle buttons
   # now on top

   # The Playlist

   frame .playedit.p -relief sunken -bd 2
   set m .playedit.p
   frame $m.t -relief sunken 
   label $m.t.label -text "Playlist:" -anchor w -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-* -relief flat -bg grey40 -fg white
   entry $m.t.filename -textvariable PlayList(filename) -relief sunken \
      -width 25 -font $PlayEdit(entryfont) \
      -bg $PlayEdit(bgcolor) -fg $PlayEdit(fghighlight) \
      -highlightcolor $PlayEdit(bgcolor) -insertbackground $PlayEdit(crcolor)
   set pl(entry) $m.t.filename
   set PlayList(filename) ""
   pack $m -side top -fill both -expand 1
   pack $m.t -side top -fill x -expand 1
   pack $m.t.label -side top -fill x -anchor w
   pack $m.t.filename -side left -fill x -expand 1


   listbox $m.playlist -relief flat \
      -yscrollcommand "$m.scroller set" -bg $PlayEdit(bgcolor) \
      -fg $PlayEdit(fgcolor) -selectbackground $PlayEdit(bgcolor) \
      -selectforeground $PlayEdit(fghighlight) -font $PlayEdit(listfont) \
      -selectborderwidth 0 -height 7 -selectmode extended
   set PlayList(listbox) $m.playlist
   scrollbar $m.scroller -command "$m.playlist yview" -width 16
   pack $m -side top -fill both -expand 1
   pack $m.t -side top -fill x
   pack $m.t.label -side top
   pack $m.t.filename -side top -fill x -expand 1
   pack $m.scroller -side left -fill y
   pack $m.playlist -side left -fill both -expand 1
   bind $m.playlist <Double-1> {modal_edit}

   DirList(update) $cdir

   if {[info exists p(playlist)]} {
      set PlayList(playlist) $p(playlist)
   }

   PlayList(update)

   return
}
